;;;;
;;;; t a g s . s c m            -- Unique tags for a poor man's
;;;;                               user-defined types.
;;;;
;;;; Copylefted under the GPL
;;;;
;;;;           Author: Francois-Rene Rideau [rideau@unice.fr]

(provide 'tags)

;;;;
;;;; Exported symbols
;;;; ----------------
;;;;   make-tag
;;;;   make-tagged
;;;;   tagged-object-maker
;;;;   tagged-object-detector
;;;;   tagged?
;;;; see documentation strings.
;;;;
;;;;
;;;; Redefined symbols
;;;; -----------------
;;;;   pair?
;;;;
;;;; Internal symbols
;;;; ----------------
;;;;   tags$$magic-tag
;;;;   tags$$pair?
;;;;   tags$$is-tag?
;;;;   tags$$tagged-object?


(define tags$$magic-tag (list '$)) ; magic marker

(define (make-tag . args)
  "creates a new, unique, tag, recognizable as such.
   as the cons of a magic tag and the given arguments"
  (append args tags$$magic-tag))
;; this implementation reduces visual overhead when displaying tagged objects

(define (make-tagged tag . fields)
  "make a tagged object of given tag and given subobjects;
  implemented as a list of the tag and the subobjects"
  (apply list tag fields))

(define (tagged-object-maker tag)
  "create a maker function for a poor man's user-type implemented as
  the cons of the type's tag and the list of initializing arguments"
  (lambda rest (apply make-tagged tag rest)))

(define (tagged-object-detector tag)
  "create a detector function for a poor man's user-type implemented as
  the cons of the type's tag and the list of initializing arguments"
  (lambda (obj) (and (tags$$pair? obj) (eq? (car obj) tag))))


;;;; Redefine pair? to not see tagged objects...

(define tags$$pair? pair?)
(define (tags$$is-tag? obj)
  (or (eq? obj tags$$magic-tag)
      (and (tags$$pair? obj)
	   (tags$$is-tag? (cdr obj)))))
(define (tags$$tagged-object? x)
  (and (tags$$pair? x) (tags$$is-tag? (car x))))
(define (pair? x)
  (and (tags$$pair? x)
       (not (tags$$is-tag? x))
       (not (tags$$tagged-object? x))))


;;;; tagged? detects tagged objects
(define tagged? tags$$tagged-object?)


; Note:
;
; We could similarly redefine car, cdr, etc, to check for a pair?,
; providing special accessors for tagged object,
; but this would only be boring and slowing everything down even more...
;
(define cons-tag (make-tag 'cons))
